# Distribution Checker
# Report Generation Routines (Report.pm)
#
# Copyright (C) 2007-2009 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
# The ptyshell tool is originally written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Report;
use strict;

use Common;

# Export symbols
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	&html
	&guess_journal_type
);
#----------------------------------------------------------------------

# Journal structure:
# $jstruct -> {
#   FILE, TESTSUITE,
#   KIND => TET|SIMPLE|TCL|PYTHON, # kind of the journal
#   STATS => { PASSED, WARNINGS, TESTS_TOTAL, NOTICES, FIPS, TESTS_EXPECTED, 
#              TESTSUITE_ERRORS, UNTESTED, FAILURES, QUELLED }, # statistical data
#
#   # @ARR elements can be of two types.
#   # Passed and untested test cases are stored as scalars (to save memory),
#   #  and their format is "$severity:$test_name",
#   #  where $severity is an empty line "" for passed test cases.
#   # Hash type is for other test result types and can contain various details.
#   ARR => [ $testcase_ok |
#            {TEST_NAME, TEST_POINT, TEST_NUM, SEVERITY, MESSAGE, LINE_NUM, COMMENT}
#           ],
#   TESTS_EXPECTED, # expected number of test cases
#   INFO_TABLES, # information from journal header
#  };

# Map of TET result codes to human-readable severity names.
our %result_code_map = (
	0 => "Passed",
	1 => "Failed",
	2 => "Unresolved",
	3 => "Unused",
	4 => "Unsupported",
	5 => "Untested",
	6 => "Uninitiated",
	7 => "Unreported",
	65 =>  "Time expired",
	101 => "Warning",
	102 => "FIP",
	103 => "Unimplemented",
	104 => "Unapproved",
);

#----------------------------------------------------------------------

our %severity_map = (
		"passed" => 'passed',
		
		"failed"       => 'failed',
		"unresolved"   => 'failed',
		"uninitiated"  => 'failed',
		"unreported"   => 'failed',
		"time expired" => 'failed',
		
		"unused"        => 'untested',
		"unsupported"   => 'untested',
		"untested"      => 'untested',
		"unimplemented" => 'untested',
		"unapproved"    => 'untested',

		"fip" => 'fip',
		
		"warning" => 'warning',
		"testsuite_warning" => 'warning',
		
		"notice"  => 'notice',
		"testsuite_notice" => 'notice',
		
		"testsuite_error"  => 'testsuite_error',
	);

#----------------------------------------------------------------------

# HTML colors used in the report.
my $dark_gray       = "#555555";
my $gray            = "#C8C8C8";
my $neutral_gray    = "#888888";
my $light_gray      = "#F0F0F0";
my $dark_blue       = "#000080";
my $dark_green      = "#006600";
my $fluorescent_green   = "#40FF40";
my $fluorescent_yellow  = "#FFFF40";
my $fluorescent_red     = "#FF6060";
my $green           = "#00A000";

#----------------------------------------------------------------------

# Problem database records
our $problem_database;

our $cert_mode = 0; # For TRIAL-USE tests

our $result_dir = undef;

#----------------------------------------------------------------------

# Loads the problem database which is used for automatic identification
# of known test suite failures, warnings and false positives.
sub load_problem_database {
	my ($file_name) = @_;

	# Open the database file.
	open FILE, $file_name
		or return error "Failed to open problem database '$file_name': $!";

	# The problem record being read.
	my $problem;

	while ( my $line = <FILE> ) {
		
		$line =~ s/^\s+//; # Remove spaces in the beginning of the line
		$line =~ s/\s+$//; # Remove trailing spaces
		
		next if $line eq "";   # Skip empty lines
		next if $line =~ /^#/; # Skip comments

		if ( !$problem ) {
			# The 'test' line starts a new test record.
			# It contains the name of the test and the test point number.
			if ($line =~ /^test\s+(.*)$/) {
				my $test_name = $1;

				# Clear the old record and prepare a clean one.
				$problem = {
					'test_name' => $test_name,
				};
				
				next;
			}
			else {
				warning "problem_db: $.:"
					." Unexpected text beyound a test block: '$line'";
			}
		}
		
		# The 'expected_severity' line indicates the severity
		# expected to be reported by the test suite.
		if ( $line =~ /^expected_severity\s+(.*)/ ) {
			my $expected_severity = lc($1);

			# Check if the expected severity is one of the predefined values.
			if ( !defined $severity_map{$expected_severity} ) {
				warning "problem_db: $.:"
					.($problem ? " ".$problem->{'test_name'}.":" : "")
					." unrecognized expected_severity '$expected_severity'."
					." Allowed values are: ".(join ", ", keys %severity_map);
			}

			$problem->{'expected_severity'} = $expected_severity;
		}
		# The 'marker' line contains a marker which can be found
		# to identify the particular problem from the messages.
		# There may be (and typically are) more markers, each on its own
		# 'marker' line.
		elsif ( $line =~ /^marker\s+(.*)/ ) {
			my $marker = $1;
			push @{$problem->{'markers'}}, $marker;
		}
		# The 'resolution' line contains the resolution of the problem.
		elsif ($line =~ /^resolution\s+(.*)/) {
			my $resolution = $1;
			$problem->{'resolution'} = $resolution;
		}
		# If the 'quell' line is present, the problem is considered harmless
		# regardless of severity. This is used for already solved bugs
		# and false positives.
		elsif ($line =~ /^quell(\s+(.*))?/) {
			my $force_severity = $2;
			$force_severity = 1 if !$force_severity;
			
			$problem->{'quell'} = $force_severity;
		}
		# The 'comment' line holds a comment for human reader
		# which is copied to the problem report. There can be more
		# comment lines.
		elsif ($line =~ /^comment\s+(.*)/) {
			my $text = $1;
			
			$problem->{'comment'} .= "\n" if defined $problem->{'comment'};
			$problem->{'comment'} .= $text;
		}
		# The 'lsb_report_id' contains the LSB problem report ID.
		elsif ($line =~ /^lsb_report_id\s+(\S+)$/) {
			$problem->{'lsb_report_id'} = $1;
		}
		# The 'lsb_resolution_id' contains the LSB resolution ID.
		elsif ($line =~ /^lsb_resolution_id\s+(\S+)$/) {
			$problem->{'lsb_resolution_id'} = $1;
		}
		# 'dup' means that problem description should be copied for
		# another test case. 
		# NOTE: 'dup' should go AFTER all records to be copied.
		# In other words, if you put something below 'dup', it won't be copied.
		elsif ($line =~ /^dup\s+(.*)$/) {
			my $dup_test_name = $1;

			my $copy = { %$problem };
			$copy->{'test_name'} = $dup_test_name;

			# Add the same problem record to the database with another key.
			push @{$problem_database->{$dup_test_name}}, $copy;
		}
		# The 'end' line finishes one problem entry.
		elsif ($line =~ /^end/) {

			# Check if this problem report has all required fields.
			unless ( $problem->{'expected_severity'} ) {
				warning "problem_db: $.:"
					.($problem ? " ".$problem->{'test_name'}.":" : "")
					." expected_severity field is required.";
			}
			
			# The problem record database is indexed by the test name.
			# Add the problem record to the database.
			push @{$problem_database->{$problem->{'test_name'}}}, $problem;
			
			$problem = undef; # Close the problem record.
		}
		# Anything else is a syntax error.
		else {
			warning "problem_db: $.:"
				.($problem ? " ".$problem->{'test_name'}.":" : "")
				." Unrecognized line: '$line'";
		}
	}
	close FILE;
	
	return 1; # OK
}
#---------------------------------------------------------------------

# Finds problem record in the database.
sub find_problem_record {
	my ( %data ) = @_;

	my $test_name  = $data{TEST_NAME};
	my $test_point = $data{TEST_POINT};
	my $severity   = $data{SEVERITY};
	my $test_messages = $data{MESSAGE};

	# Some automatically generated problem reports have no test name or point;
	# these are not meant to be searched in the database but directly reported
	# to the user.
	return undef if !$test_name;

	# TODO: bind each problem record to version(s) of test suite.
	# TODO: regexp in matcher.

	# The basic record key is the name of the test and the test point number.
	my $record_key = $test_name;
	$record_key .= " " . $test_point if defined $test_point && $test_point ne "";
	
	# Get all records matching the key
	( $problem_database ) or return undef;
	my $found_records = $problem_database->{$record_key};
	
	return undef if !$found_records || !@$found_records; # Nothing found;
	
	# Try finding the problem record.
	PROBLEM_RECORD:
	foreach my $problem_record ( @$found_records ) {

		# Check if the severity matches.
		my $expected_severity = $problem_record->{'expected_severity'};
		next PROBLEM_RECORD  unless lc($severity) eq lc($expected_severity);
			
		# Look which distinguishing markers the problem should have.
		my $markers = $problem_record->{'markers'};
		
		# Check if the markers are found in the test messages.
		foreach my $marker ( @$markers ) {
			unless ( index($test_messages, $marker) >= 0 ) {
				# The message doesn't match to marker.
				next PROBLEM_RECORD; # Try another problem record.
			}
		}
		
		# All the markes match.
		# Yup, this is the problem report we looked for.
		return $problem_record;
	}

	# Nothing was found.
	return undef;
}
#---------------------------------------------------------------------

# Escape HTML entities
sub html {
	my ($text) = @_;
	
	( defined $text ) or return error "Undefined parameter";
	
	$text =~ s/&/&amp;/g;

	$text =~ s/</&lt;/g;
	$text =~ s/>/&gt;/g;
	$text =~ s/"/&#34;/g;
	
	#$text =~ s/([\x80-\xFF])/'&#'.ord($1).';'/ge;
	$text =~ s/[\x00-\x08\x0B\x0C\x0E-\x1F]/&#9072;/g;
	return $text;
}

# Escape HTML entities except for text in {noreplace}{/noreplace}
sub html_excl {
	my ($text) = @_;
	
	( defined $text ) or return undef;
	
	my $res = "";
	
	while ( $text =~ s/^(.*?){noreplace}//s ) {
		$res .= html($1);
		if ( $text =~ s/^(.*?){\/noreplace}//s ) {
			$res .= $1; # without escaping
		} else {
			$res .= "{noreplace}";
			die $text if DEBUG; # DBG:
		}
	}
	
	$res .= html($text);
	
	return $res;
}
#----------------------------------------------------------------------

sub ErrMessage {
	my ($error, $title) = @_;
	
	my $res = "";
	
	if ( is_err($error) ) {
		$res .= '<table class="errmsg" style="padding: 0pt 15pt 0pt 0pt;"><tr><td>'."\n";
		$res .= html($title) if defined $title;
		$res .= '<ul>';
		local *error_li = sub {
			my ($err) = @_;
			my $res2 = "";
			if ( $err->{-arr} ) {
				return join "\n", map error_li($_), @{$err->{-arr}};
			}
			
			my $txt = html($err->{-text});
			$txt =~ s/\n/<br \/>/g;
			
			$res2 .= '<li>';
			$res2 .= $txt;
			if ( defined $err->{-prev} ) {
				$res2 .= "\n".'<ul>';
				$res2 .= error_li($err->{-prev});
				$res2 .= '</ul>'."\n";
			}
			$res2 .= '</li>';
			return $res2;
		};
		$res .= error_li($error);
		$res .= '</ul>';
		$res .= '</td></tr></table>';
	} else {
		$res .= '<pre class="errmsg">';
		$res .= html($title) if defined $title;
		$res .= html($error);
		$res .= '</pre>';
	}
	
	return $res;
}
#----------------------------------------------------------------------

# Makes a hyperlink to a file in the "results" subdir.
sub local_href {
	my ($href, $content, $more) = @_;
	
	( defined $href ) or return error "Undefined parameter";
	
	$content = $href if !defined $href;
	
	if ( !defined $more ) {
		$more = 'target="_blank"';
	}
	
	my $res = "";
	
	my $link = $href;
	if ( $link !~ /^\// && defined $result_dir ) {
		$link = $result_dir.'/'.$link;
	}
	if ( $link =~ /^\// ) {
		$link =~ s/\?/\&/; # Replace GET separator.
		$res .= '<a class="view_srv" '.$more.' href="'.html('/get.pl'.$link).'">'
				.$content.'</a>';
	}
	
	if ( defined $result_dir && $href =~ /^\// ) {
		if ( substr($href,0,length($result_dir)) eq $result_dir ) {
			$href = substr($href,length($result_dir));
			$href =~ s{^/+}{};
		}
	}
	if ( $href !~ /^\.*\// ) {
		$href = "./".$href;
	}
	$res .= '<a class="view_as_is" '.$more.' href="'.html($href).'">'.$content.'</a>';
	
	return $res;
}
#----------------------------------------------------------------------

sub ParseJournal {
	my ($jstruct) = @_;
	
	( defined $jstruct ) or return error "Undefined parameter";
	( ref($jstruct) eq 'HASH' ) or return error "Wrong parameter";
	
	my $kind = $jstruct->{KIND} or return error "Journal kind undefined";
	my $file = $jstruct->{FILE} or return error "Journal file undefined";
	
	if ( $kind eq 'TET' ) {
		is_ok ParseTetJournal( $jstruct )
			or return error "Failed to parse TET journal '$file'", $Error::Last;
	}
	elsif ( $kind eq 'SIMPLE' ) {
		is_ok ParseSimpleJournal( $jstruct )
			or return error "Failed to parse Simple journal '$file'", $Error::Last;
	}
	elsif ( $kind eq 'TCL' ) {
		is_ok ParseTclJournal( $jstruct )
			or return error "Failed to parse TCL journal '$file'", $Error::Last;
	}
	elsif ( $kind eq 'PYTHON' ) {
		is_ok ParsePythonJournal( $jstruct )
			or return error "Failed to parse Python journal '$file'", $Error::Last;
	}
	else {
		return error "Unknown journal kind: '$kind'";
	}
	
	# Calculate statistics
	my $stats = {
			TESTS_TOTAL => 0,
			PASSED      => 0,
			FAILURES    => 0,
			WARNINGS    => 0,
			QUELLED     => 0,
			NOTICES     => 0,
			UNTESTED    => 0,
			FIPS        => 0,
			TESTS_EXPECTED => undef,
			TESTSUITE_ERRORS => 0,
		};
	
	foreach my $obj ( @{$jstruct->{ARR}} ) {
		$stats->{TESTS_TOTAL}++;
		
		if ( ref($obj) eq "" ) {
			my $severity;
			my $testcase;
			if ( $obj =~ /^([^:]*):(.*)$/ ) {
				$severity = $1;
				if ( $severity eq "" ) { $severity = "passed" }
				$testcase = $2;
			}
			else {
				warning "Wrong jstruct entry: '$obj'";
				next;
			}
			
			my $gen_severity = $severity;
			$gen_severity = $severity_map{$severity} if defined $severity_map{$severity};
			
			if ( $gen_severity eq 'passed' ) {
				$stats->{PASSED}++;
			}
			elsif ( $gen_severity eq 'untested' ) {
				$stats->{UNTESTED}++;
			}
			else {
				warning "Unknown severity code: '$severity'";
				next;
			}
			
			next;
		}
		# else:
		
		if ( ref($obj) ne "HASH" ) {
			warning "Wrong entry type in jstruct";
			next;
		}
		
		my $severity = $obj->{SEVERITY};
		
		my $gen_severity = $severity;
		$gen_severity = $severity_map{$severity} if defined $severity_map{$severity};
		
		$stats->{TESTS_TOTAL}-- if $severity =~ /^testsuite_/; # Not a real test case
		
		if ( $severity =~ /(quelled|waived)/ ) { # Contains "quelled" or "waived"
			$stats->{QUELLED}++; # Increment the quelled tests count statistic.
		}
		elsif ( $gen_severity eq "testsuite_error" ) {
			$stats->{TESTSUITE_ERRORS}++;
		}
		elsif ( $gen_severity eq "failed" ) {
			$stats->{FAILURES}++; # Increment the number of failures reported.
		}
		elsif ( $gen_severity eq "warning" ) {
			$stats->{WARNINGS}++; # Increment the number of warnings reported.
		}
		elsif ( $gen_severity eq "fip" ) {
			$stats->{FIPS}++; # Increment the number of FIPs reported.
		}
		elsif ( $gen_severity eq "notice" ) {
			$stats->{NOTICES}++; # Increment the number of notices reported.
		}
		else {
			warning "Unhandled severity: '$severity' {$gen_severity}";
			$stats->{TESTSUITE_ERRORS}++;
		}
	}
	
	if ( defined $jstruct->{TESTS_EXPECTED} ) {
		$stats->{TESTS_EXPECTED} = $jstruct->{TESTS_EXPECTED};
		
		if ( $stats->{TESTS_EXPECTED} < 0 ) {
			$stats->{TESTS_EXPECTED} = $stats->{TESTS_TOTAL} + (-$stats->{TESTS_EXPECTED});
		}
	}

	if ( $stats->{TESTS_TOTAL} == 0 ) {
		wrong_structure ( { JS => $jstruct }, "No tests in journal", 
				"The test journal contains no test records.", "testsuite_error" );
		$stats->{TESTSUITE_ERRORS}++;
	}
	
	$jstruct->{STATS} = $stats;
	
	return $jstruct;
}
#----------------------------------------------------------------------

sub ParseTetJournal {
	my ($jstruct) = @_;
	
	my $file = $jstruct->{FILE} or return error "Journal file undefined";
	
	open JOURNAL_FH, $file
		or return error "Can't open file for reading: '$file': $!";
	
	# Flags indicating that we are inside a some block.
	my $inside_test_case_controller = 0; # between code 0 and code 900
	
	my $test_comment = undef;
	my $test_messages = undef;
	my $test_points_in_this_case;
	my $line_number = 0;
	my %config_vars = (); # For lines with Code 30 -- Configuration Variables.
	
	my $context = { JS => $jstruct };
	
	$context->{TEST_NAME} = undef;
	$context->{LINE_NUM} = 0;
	
	# Read the journal line by line.
	while ( my $line = <JOURNAL_FH> ) {
		chomp $line;
		
		#$context->{LINE_TEXT} = $line;
		$context->{LINE_NUM}++;
		
		# Each journal line consists of three parts, separated by
		# a vertical bar:
		#
		# 1. A numeric line code, determining the meaning of the line.
		# 2. Additional data, format depends on line code.
		# 3. Additional text (message for the user, or some comment).
		# e.g. '400|0 1 1 20:04:00|IC Start'
		
		my ($line_code, $line_data, $line_text) = split /\|/, $line, 3;
		
		## Determine the type of the line and process it appropriately.
		
		if ($line_code == 0) {
		# Code 0 -- Test Case Controller Start.
			# 0|version time date|who
			my ($test_engine_version, $test_start_time, $test_start_date);
			if ( $line_data =~ /^(\S*)\s+(\S*)\s+(\S*)$/ ) {
				$test_engine_version = $1;
				$test_start_time = $2;
				$test_start_date = $3;
			}
			else {
				wrong_structure ( $context, -NAME, "Wrong syntax",
					"Unrecognized entry in journal." );
				next;
			}
			
			if ( $inside_test_case_controller ) {
				if ( 0 ) { # TODO: some switch
					wrong_structure ( $context, -NAME, "Wrong syntax",
						"TCC Start before the previous TCC block has been closed." );
				}
			}
			$inside_test_case_controller = 1;
			
			# Print the test engine version, time and date to the output.
			my $info_table = {};
			$info_table->{TITLE} = "Information about the test suit from the journal";
			push @{$info_table->{ROWS}}, {LEFT => "Test Engine Version", RIGHT => $test_engine_version};
			push @{$info_table->{ROWS}}, {LEFT => "Test Started At", RIGHT => $test_start_time." ".$test_start_date};
			push @{$info_table->{ROWS}}, {LEFT => "Test Command Line", RIGHT => $line_text};
			
			push @{$jstruct->{INFO_TABLES}}, $info_table;
		}
		elsif ($line_code == 5) {
		# Code 5 -- Local System Information.
			# The line structure is:
			# 5|sysname nodename release date machine|text
			my ($system_name, $node_name, $release, $date, $machine);
			if ( $line =~ /^5\|(\S+)\s+(\S+)\s+(\S+)\s+#(.+)\s+(\S+)\|/ ) {
				$system_name = $1;
				$node_name = $2;
				$release = $3;
				$date = $4;
				$machine = $5;
			}
			else {
				wrong_structure ( $context, -NAME, "Wrong syntax",
					"Unrecognized entry in journal." );
				next;
			}
			
			# Save the local system data.
			my $info_table = {};
			$info_table->{TITLE} = "Information about the system under test from the journal";
			push @{$info_table->{ROWS}}, {LEFT => "System Name", RIGHT => $system_name};
			push @{$info_table->{ROWS}}, {LEFT => "Node Name", RIGHT => $node_name};
			push @{$info_table->{ROWS}}, {LEFT => "Release", RIGHT => $release.' #'.$date};
			push @{$info_table->{ROWS}}, {LEFT => "Machine", RIGHT => $machine};
			
			push @{$jstruct->{INFO_TABLES}}, $info_table;
		}
		elsif ($line_code == 10) {
		# Code 10 -- Test Case Start.
			# The line structure is:
			# 10|activitynum test_name hh:mm:ss|list of invocable components
			# * Timestamp is omitted in perl test reports. (Bug?)
			my ($activitynum, $testname_param, $timestamp);
			if ($line_data =~ /^(\d+)\s+(\S+)\s*(\S*)$/) {
				$activitynum = $1;
				$testname_param = $2;
				$timestamp = $3;
			}
			else {
				wrong_structure ( $context, -NAME, "Wrong syntax",
					"Unrecognized entry in journal." );
				next;
			}
			
			if ( defined $context->{TEST_NAME} ) {
				# The previous test case was not properly finished;
				# issue a warning.
				wrong_structure ( $context, -NAME, "Previous TC hasn't finished", 
						"Test case ($testname_param) starts before"
							." the previous one (".$context->{TEST_NAME}.") has finished"
					);
			}
			
			# Set the test name and number.
			$context->{TEST_NUM} = $activitynum;
			$context->{TEST_NAME} = $testname_param;
			
			# Reset the test point number and counter.
			$context->{TEST_POINT} = undef;
			$test_points_in_this_case = 0;
		}
		elsif ($line_code == 15) {
			# Code 15 -- Test Case Manager Start.
		}
		elsif ($line_code == 20) {
			# Code 20 -- Local/Remote System Configuration Start.
		}
		elsif ($line_code == 30) {
		# Code 30 -- Configuration Variables.
			# Format: 30||VARIABLE=value
			if ($line =~ /^30\|\|([^=]+)=(.*)/) {
				my $varname = $1;
				my $value = $2;
				
				$config_vars{$varname} = $value;
			}
		}
		elsif ($line_code == 40) {
		# Code 40 -- Configuration End.
		}
		elsif ($line_code == 50) {
		# Code 50 -- Test Case Controller Message.
			( $inside_test_case_controller )
				or wrong_structure ( $context, "Wrong syntax",
					"A TCC Message encountered beyond a TCC block." );
			
			# Format: 50||Message
			process_test_point_result ( %$context,
						SEVERITY => "testsuite_error",
						COMMENT => "Test case controller reports a problem:",
						MESSAGE => $line_text, );
		}
		elsif ($line_code == 70) {
		# Code 70 -- Scenario Information.
			
			# LSB-runtime-test only:
			# for the .os sections need to double expected counts to allow
			# for the macro versions.
			if ( $line =~ /^70\|.*total tests in (ANSI|POSIX|LSB|PTHR).os ([0-9]+)/ ) {
				$jstruct->{TESTS_EXPECTED} += $2 * 2;
			}
			elsif ( $line =~ /^70\|.*total tests in ([A-Za-z.0-9_-]+) ([0-9]+)/ ) {
				$jstruct->{TESTS_EXPECTED} += $2;
			}
			elsif ( $line =~ /total\s+tests\s+in/ ) {
				wrong_structure ( $context, "Wrong syntax",
					"Unrecognized entry in journal." );
			}
		}
		elsif ($line_code == 80) {
		# Code 80 -- Test Case End.
			if ( !defined $context->{TEST_NAME} ) {
				# Test case end found when no test case was started;
				# issue a warning.
				wrong_structure ( $context, "Unexpected TC End",
						"A Test Case End entry found when no test case seems to be running" );
			}

			# Warn if the test case seems to be empty.
			if ( $test_points_in_this_case == 0 ) {
				wrong_structure ( $context, "No test points",
						"The test case contains no test points" );
			}
			
			# Check the test point returned any result
			if ( defined $context->{TEST_POINT} ) {
				wrong_structure ( $context, "No result",
						"The test case returned no result", "unreported" );
			}
			
			# Clear the test name, number, messages, and others.
			$context->{TEST_NAME} = undef;
			$context->{TEST_NUM} = 0;
			$test_comment = undef;
			$test_messages = undef;
			$test_points_in_this_case = 0;
		}
		elsif ($line_code == 90) {
		# Code 90 -- User Abort.
			process_test_point_result ( %$context,
						SEVERITY => "testsuite_error",
						RESOLUTION  => "Interrupted by the user",
						COMMENT => "Test suite execution was interrupted by the user"
					);
		}
		elsif ($line_code == 100) {
			# Code 100 -- Captured Output.
			if ( defined $line_text ) {
				$line_text =~ s/\s+/ /g;
				$test_messages .= "| ".$line_text."\n";
			}
		}
		elsif ($line_code == 110) {
			# Code 110 -- Build Start.
		}
		elsif ($line_code == 130) {
			# Code 130 -- Build End.
		}
		elsif ($line_code == 200) {
		# Code 200 -- Test Purpose Start.
			# The structure of the line is as follows:
			# 200|activitynum pointnum hh:mm:ss|comment
			unless ( $line =~ /^\d*\|\d*\s*(\d*)/ ) {
				wrong_structure ( $context, "Wrong syntax", "Unrecognized entry in journal." );
				next;
			}
			my $test_point = $1;
			
			# Check if the previous test point was properly finished.
			if ( defined $context->{TEST_POINT} ) {
				if ( $context->{TEST_POINT} == $test_point ) {
					# Some test suites have a strange habit of re-starting
					# a test point.
					
					# TODO: some switch
					if ( 0 ) {
						wrong_structure ( $context, "Restarted in the middle",
							"The test was restarted in the middle", "testsuite_notice");
					}
				}
				else {
					# Notify about a disappeared test.
					wrong_structure ( $context, "Previous test point hasn't finished",
							"A test point starts before the previous one"
							." (".$context->{TEST_POINT}.") has finished" );
				}
			}
			
			
			# Change the test point number.
			$context->{TEST_POINT} = $test_point if defined $test_point && $test_point ne "";
			
			# Increment the number of test points.
			$test_points_in_this_case++;
			
			# Clear the messages.
			$test_messages = undef;
			
			if ( $line_text && $line_text !~ "TP Start" ) {
				$test_comment = "Test Purpose Comment: ".$line_text;
			} else {
				$test_comment = undef;
			}
		}
		elsif ($line_code == 220) {
		# Code 220 -- Test Purpose Result.
			# The structure of the line is as follows:
			# 220|activitynum pointnum result time|comment
			my ($activitynum, $pointnum, $result_code, $timestamp);
			if ($line_data =~ /^(\d+)\s+(\d+)?\s+(\d+)\s+(\S+)$/) {
				$activitynum = $1;
				$pointnum = $2;
				$result_code = $3;
				$timestamp = $4;
			}
			else {
				wrong_structure ( $context, "Wrong syntax",
					"Unrecognized entry in journal." );
				next;
			}
			
			# Print a warning if the test purpose seems to never been started.
			if ( !defined $context->{TEST_POINT} ) {
				wrong_structure ( $context, "Unexpected test result entry",
						"A test result entry found but the test point had not started" );
				$context->{TEST_POINT} = $pointnum;
			}
			
			# Print a warning if the test purpose number is different from
			# what we expected.
			if ( defined $pointnum && $context->{TEST_POINT} != $pointnum) {
				wrong_structure ( $context, "Unexpected TP number",
						"Test purpose number '$pointnum' differs"
						." from '".$context->{TEST_POINT}."' at the beginning of the test." );
			}
			
			# Find the meaning of the result code (severity).
			if ( !defined $result_code_map{$result_code} ) {
				wrong_structure ( $context, "Unknown result code",
						"Unrecognized test result code ($result_code)" );
				next;
			}
			
			my $severity = $result_code_map{$result_code};
			
			# Process the result.
			process_test_point_result ( %$context,
						SEVERITY => lc $severity,
						MESSAGE => $test_messages,
						COMMENT => $test_comment,
					);
			
			# Clear the test point number.
			$context->{TEST_POINT} = undef;
		}
		elsif ($line_code == 310) {
			# 310 -- Clean Start.
		}
		elsif ($line_code == 320) {
			# 320 -- Clean End.
		}
		elsif ($line_code == 400) {
			# 400 -- Invocable Component Start.
		}
		elsif ($line_code == 410) {
			# 410 -- Invocable Component End.
			
			# Check the test point returned any result
			if ( defined $context->{TEST_POINT} ) {
				wrong_structure ( $context, "No result",
						"The test case returned no result", "unreported", $test_messages );
				$context->{TEST_POINT} = undef;
			}
		}
		elsif ($line_code == 510) {
		# 510 -- Error reported by the underlying test suite.
			# Format: 510|testnum|Message
			process_test_point_result ( %$context,
						SEVERITY => "testsuite_error",
						COMMENT => "Error reported by the test suite:",
						MESSAGE => $line_text, );
		}
		elsif ($line_code == 520) {
		# 520 -- Message from the underlying test suite.
			
			if ( defined $line_text ) {
				# Coalesce consequent rows of spaces in the message
				# to a single space. This improves matching with markers.
				$line_text =~ s/\s+/ /g;
				
				# Add the message to the accumulated messages.
				$test_messages .= $line_text."\n";
			}
		}
		elsif ($line_code == 600) {
			# 600 -- Parallel Start.
		}
		elsif ($line_code == 620) {
			# 620 -- Parallel End.
		}
		elsif ($line_code == 630) {
			# 630 -- Implied Sequential Start.
		}
		elsif ($line_code == 640) {
			# 640 -- Implied Sequential End.
		}
		elsif ($line_code == 700) {
			# 700 -- Repeat Start.
		}
		elsif ($line_code == 720) {
			# 720 -- Repeat End.
		}
		elsif ($line_code == 730) {
			# 730 -- Timed Loop Start.
		}
		elsif ($line_code == 740) {
			# 740 -- Timed Loop End.
		}
		elsif ($line_code == 750) {
			# 750 -- Random Start.
		}
		elsif ($line_code == 760) {
			# 760 -- Random End.
		}
		elsif ($line_code == 800) {
			# 800 -- Remote Start.
		}
		elsif ($line_code == 820) {
			# 820 -- Remote End.
		}
		elsif ($line_code == 830) {
			# 830 -- Distributed Start.
		}
		elsif ($line_code == 840) {
			# 840 -- Distributed End.
		}
		elsif ($line_code == 900) {
			# 900 -- Test Case Controller End.
			if ( 0 ) {
				( $inside_test_case_controller )
					or wrong_structure ( $context, "Wrong syntax",
						"TCC End encountered beyond a TCC block." );
			}
			$inside_test_case_controller = 0;
		}
		else {
			# Other lines are notified to the user and otherwise ignored.
			wrong_structure ( $context, "Unrecognized entry",
					"Unrecognized journal entry" );
		}
	}
	
	# Check the journal was closed correctly
	if ( $inside_test_case_controller ) {
		wrong_structure ( $context, -NAME, "Wrong syntax",
					"TCC block wasn't closed. The journal may be incomplete/broken." );
	}
	if ( defined $context->{TEST_NAME} ) {
		wrong_structure ( $context, "Wrong syntax", 
				"Test Case block wasn't closed (".$context->{TEST_NAME}."). The journal may be incomplete/broken." );
	}
	
	# Close the file
	close JOURNAL_FH;
	
	# Save the configuration info.
	if ( %config_vars ) {
		my $info_table = {};
		$info_table->{TITLE} = "Configuration information from the journal";
		
		for my $varname qw(VSX_NAME LSB_VERSION LSB_MODULES TET_VERSION XT_FONTPATH) {
			if ( defined $config_vars{$varname} ) {
				push @{$info_table->{ROWS}}, 
						{ LEFT => $varname, RIGHT => $config_vars{$varname} };
				delete $config_vars{$varname};
			}
		}
		
		push @{$jstruct->{INFO_TABLES}}, $info_table;
	}
}
#----------------------------------------------------------------------

sub ParseSimpleJournal {
	my ($jstruct) = @_;
	
	my $file = $jstruct->{FILE} or return error "Journal file undefined";
	
	open JOURNAL_FH, $file
		or return error "Can't open file for reading: '$file': $!";
	
	# Flags indicating that we are inside a some block.
	my $finished = 0;
	
	my $test_description = undef;
	my $test_comment = undef;
	my $test_message = undef;
	
	my $line_number = 0;

	my $context = { JS => $jstruct };

	$context->{TEST_NAME} = undef;

	# Read the journal line by line.
	while ( my $line = <JOURNAL_FH> ) {
		chomp $line;
		
		$line_number++;
		
		$line =~ s/^\s+//; # Trim spaces
		$line =~ s/\s+$//;
		
		next if $line eq "";   # Skip empty lines
		next if $line =~ /^#/; # Skip comments
		
		$context->{LINE_NUM}  = $line_number;
		#$context->{LINE_TEXT} = $line;
		
		my ($linetype, $value);
		
		# Format: "LINETYPE: some text"
		if ( $line =~ /^([A-Z_\-]+)\s*(?:\:\s*(.*))?$/ ) {
			$linetype = $1;
			$value = $2;
		}
		else {
			# Lines that do not conform to this format are reported
			# to the user and skipped.
			wrong_structure( $context, -NAME, "Wrong syntax",
					"Unrecognized entry in journal." );
			next;
		}
		
		if ( $linetype eq 'TEST') {
			# Ignore
		}
		elsif ( $linetype eq 'TESTS_EXPECTED' ) {
			$jstruct->{TESTS_EXPECTED} = $value;
		}
		elsif ( $linetype eq 'TESTCASE' ) {
			if ( defined $context->{TEST_NAME} ) {
			# Previous test case wasn't closed
				wrong_structure_line( $context, "Unclosed test case",
						"New testcase '$value' was started"
						." before the previous one has closed: '".$context->{TEST_NAME}."'." );
			}
			
			$context->{TEST_NAME} = $value;
			$test_description = undef;
			$test_comment = undef;
			$test_message = undef;
		}
		elsif ( $linetype =~ /^(DESCR|COMMENT|MESSAGE)$/ ) {
			if ( !defined $context->{TEST_NAME} ) {
				wrong_structure ( $context, "Unexpected line",
					"$linetype beyond a test block." );
				next;
			}
			
			if ( $linetype eq 'DESCR' ) {
				$test_description .= $value."\n";
			}
			elsif ( $linetype eq 'COMMENT' ) {
				$test_comment .= $value."\n";
			}
			elsif ( $linetype eq 'MESSAGE' ) {
				$test_message .= $value."\n";
			}
			else {
				wrong_structure ( $context, "Missed line type",
					"Missed line type: '$linetype'." );
			}
		}
		elsif ( $linetype eq 'RESULT' ) {
			# Process the result.
			process_test_point_result ( %$context,
						SEVERITY => lc $value,
						DESCRIPTION => $test_description,
						COMMENT => $test_comment,
						MESSAGE => $test_message,
					);
			
			$context->{TEST_NAME} = undef;
			$test_description = undef;
			$test_comment = undef;
			$test_message = undef;
		}
		elsif ( $linetype eq 'FINISHED' ) {
		# This record indicates the journal was closed correctly
			$finished = 1;
		}
		else {
			wrong_structure( $context, "Wrong syntax",
					"Unhandled line type: '$linetype'." );
		}
	}
	
	if ( defined $context->{TEST_NAME} ) {
	# Unclosed test case
		wrong_structure ( $context, "Unclosed test case",
					"Unclosed test case: '".$context->{TEST_NAME}."'." );
	}
	
	if ( !$finished ) {
	# Unclosed test case
		wrong_structure ( $context, -NAME, "Incomplete journal",
					"The journal wasn't closed correctly." );
	}
}
#----------------------------------------------------------------------

# The Tcl-style journal parser function.
sub ParseTclJournal {
	my ($jstruct) = @_;
	
	my $file = $jstruct->{FILE} or return error "Journal file undefined";
	
	open JOURNAL_FH, $file
		or return error "Can't open file for reading: '$file': $!";
	
	# Flags indicating that we are inside a some block.
	my $finished = 0;
		
	my $test_messages = undef;
	
	my $line_number = 0;
	
	my $context = { JS => $jstruct };

	$context->{TEST_NAME} = undef;
	
	# Read the journal line by line.
	while ( my $line = <JOURNAL_FH> ) {
		chomp $line;
		
		$line_number++;
		
		# Do not trim spaces at the beginning!
		$line =~ s/\s+$//; # Trim spaces at the end of the line.
		
		# Don't skip empty lines
		
		$context->{LINE_NUM}  = $line_number;
		#$context->{LINE_TEXT} = $line;
		
		# A test record message ends with a line with a rough format
		# '==== testname result'
		if ( $line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*FAILED/ ) {

			if ( !defined $context->{TEST_NAME} ) {
				# Oops, we have missed something...
				wrong_structure ( $context, "Failure",
					"End of a test record found but no start was seen." );
			}

			# Process the result.
			process_test_point_result ( %$context,
						SEVERITY => "failed",
						MESSAGE => $test_messages,
					);

			# Clear the previous test messages.
			$test_messages = undef;

			# Finish the test record.
			$context->{TEST_NAME} = undef;
		}
		# A failure report starts with a line with a rough format
		# '==== testname errormessage result'.
		elsif ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*(.+)\s*FAILED/) {

			# Load the test name.
			$context->{TEST_NAME} = $1;

			# Add the error message to the test messages.
			$test_messages .= $2."\n";
		}
		# A line that contains only the name of the test (which ends
		# with '.test', marks a passed test.
		# There could be some garbage at the beginning of the line (from a previous test).
		elsif ($line =~ /^(.*\s+)?([a-zA-Z0-9_\-]+)\.test$/) {
			$context->{TEST_NAME} = $2;
			$test_messages = undef;
			
			# Process the result.
			process_test_point_result ( %$context,
						SEVERITY => "passed",
					);
			$context->{TEST_NAME} = undef;
		}
		# Anything else?
		else {
			if ( defined $context->{TEST_NAME} ) {
				# We are inside a test point result record, so it will be
				# a test message.
				$test_messages .= $line."\n";
			}
			else {
				# Anything else is better ignored (there is too much noise
				# for considering unrecognized lines to be errors).
			}
		}
	}
}

#----------------------------------------------------------------------

# The Python-style journal parser function.
sub ParsePythonJournal {
	my ($jstruct) = @_;
	
	my $file = $jstruct->{FILE} or return error "Journal file undefined";
	
	open JOURNAL_FH, $file
		or return error "Can't open file for reading: '$file': $!";
	
	my $severity = undef;
	my $test_messages = undef;
	my $line_number = 0;
	my $finished = 0;
	
	my $context = { JS => $jstruct };
	
	local *check_prev_tc = sub {
		my ($new_test_name) = @_;
		
		if ( defined $context->{TEST_NAME} && $context->{TEST_NAME} ne $new_test_name ) {
			$context->{LINE_NUM} = $context->{TEST_LINE_NUM};
			# Save the previous test case
			process_test_point_result ( %$context,
					SEVERITY => $severity,
					MESSAGE => $test_messages,
				);
			$context->{LINE_NUM} = $line_number;
		}
	};
	
	# Read the journal line by line.
	while (my $line = <JOURNAL_FH>) {
		chomp $line;
		
		$line_number++;
		
		# Do not trim spaces at the beginning!
		$line =~ s/\s+$//; # Trim spaces at the end of the line.
		
		# Don't skip empty lines
		
		#$context->{LINE_TEXT} = $line;
		$context->{LINE_NUM}  = $line_number;

		if ( $line =~ /^(test_[a-zA-Z0-9_]+)$/ ) {
		# Looks like a test name
			my $test_name = $1;
			check_prev_tc($test_name);
			
			$context->{TEST_NAME} = $test_name;
			$context->{TEST_LINE_NUM} = $line_number;
			$severity = "passed";
			$test_messages = undef;
		}
		# A line in form 'testname skipped -- ...' marks a test
		# that was skipped, and the end of the line contains
		# the description why.
		elsif ( $line =~ /^(test_[a-zA-Z0-9_]+)\s*skipped\s*--\s*(.*)$/ ) {
			my $test_name = $1;
			my $msg = $2;
			check_prev_tc($test_name);
			
			$context->{TEST_NAME} = $test_name;
			$context->{TEST_LINE_NUM} = $line_number;
			$severity = "unsupported";
			$test_messages = $msg;
		}
		# A line in form 'test testname failed -- ...' marks a failed test.
		elsif ( $line =~ /^test\s*([a-zA-Z0-9_]+)\s*(failed|crashed)\s*--\s*(.*)$/ ) {
			my $test_name = $1;
			my $new_severity = $2;
			my $msg = $3;
			check_prev_tc($test_name);
			
			$context->{TEST_NAME} = $test_name;
			$context->{TEST_LINE_NUM} = $line_number;
			$severity = lc $new_severity;
			$severity = "failed" if $severity eq "crashed";
			$test_messages = $msg;
		}
		else {
			if ( $line =~ /^(\d+)\s+tests? OK/ ) {
			# End of the journal
				check_prev_tc("");
				
				$context->{TEST_NAME} = undef;
				$finished = 1;
				next;
			}
			if ( $line =~ /^((\d+)\s+skips unexpected)/ ) {
				my $tmp = $1;
				my $skips_num = $2;
				check_prev_tc("");
				
				$jstruct->{TESTS_EXPECTED} = -$skips_num; # Negative value means SKIPS_UNEXPECTED
					# TESTS_EXPECTED = TESTS_TOTAL - (-SKIPS_UNEXPECTED);
				
				$context->{TEST_NAME} = $tmp;
				$context->{TEST_LINE_NUM} = $line_number;
				$test_messages = $line;
				$severity = "testsuite_warning";
				next;
			}
			
			if ( defined $context->{TEST_NAME} ) {
				# Just append this line to the test record messages
				# and continue by the next line.
				$test_messages .= "\n".$line;
			} else {
				check_prev_tc("");
				
				# Other lines are ignored.
			}
		}
	}
	
	check_prev_tc("");
	
	if ( !$finished ) {
		wrong_structure ($context, -NAME, "Journal wasn't finished");
	}
}
#----------------------------------------------------------------------

sub wrong_structure {
	my $context = shift;
	my %data = %$context;
	if ( defined $_[0] && ref($_[0]) eq '' && $_[0] eq '-NAME' ) {
		delete $data{TEST_NAME};
		delete $data{TEST_POINT};
		shift; # -NAME
	}
	my $resolution = shift || "Wrong structure";
	my $comment = shift || $resolution;
	my $severity = shift || "testsuite_warning";
	my $message = shift;
	
	$comment = 'Line '.$data{LINE_NUM}.': '.$comment if defined $data{LINE_NUM};
	
	process_test_point_result ( %data,
				SEVERITY    => lc $severity,
				RESOLUTION  => $resolution,
				COMMENT     => $comment,
				MESSAGE     => $message,
			);
}
#----------------------------------------------------------------------

sub process_test_point_result {
	my ( %data ) = @_;
	
	( defined $data{SEVERITY} ) or return error "undefined severity";
	
	my $gen_severity = ( $severity_map{$data{SEVERITY}} or $data{SEVERITY} );
	my $severity_ok = ( $gen_severity eq "passed" || $gen_severity eq "untested" );
	
	my $problem_record = undef;
	
	unless ( $severity_ok ) {
		# Search for the problem record in the database.
		$problem_record = find_problem_record( %data );
		if ( !is_ok($problem_record) ) {
			warning $Error::Last;
			$problem_record = undef;
		}
	}
	
	if ( $problem_record ) {
		
		$data{PROBLEM_INFO} = $problem_record;
		
		if ( $problem_record->{'quell'} ) {
			if ( defined $severity_map{ $problem_record->{'quell'} } ) {
			# If 'quell' contains severity name, that severity will be forced.
				$data{SEVERITY} = lc $problem_record->{'quell'};
			} else {
				# Add the "(quelled)" keyword to the severity if the problem is quelled.
				# This means it can be safely ignored.
				if ( $problem_record->{'lsb_report_id'} ) {
					$data{SEVERITY} .= " (waived)";
				} else {
					$data{SEVERITY} .= " (quelled)";
				}
			}
		}
		
		$data{RESOLUTION} = $problem_record->{"resolution"} if $problem_record->{"resolution"};
		
		if ( $data{RESOLUTION} =~ /^\s*$/ ) {
			# If resolution is empty, put "..." in its place to indicate,
			# that some information about the problem is available.
			$data{RESOLUTION} = "...";
		}
	}

	# The severity may have changed, update the state
	$gen_severity = ( $severity_map{$data{SEVERITY}} or $data{SEVERITY} );
	$severity_ok = ( $gen_severity eq "passed" || $gen_severity eq "untested" );

	my $testcase = $data{TEST_NAME};
	$testcase = "***" if !defined $testcase; # testsuite error
	
	$testcase .= " ".$data{TEST_POINT} if defined $data{TEST_POINT};
	
	if ( $severity_ok ) {
		# For severity_ok save the testcase only (scalar)
		if ( $data{SEVERITY} eq "passed" ) {
			push @{$data{JS}->{ARR}}, ":".$testcase;
		} else {
			push @{$data{JS}->{ARR}}, $data{SEVERITY}.":".$testcase;
		}
	} else {
		# For problems save the entire record (hashref)
		my $record = { %data };
		delete $record->{JS};
		
		push @{$data{JS}->{ARR}}, $record;
	}
}
#----------------------------------------------------------------------

sub print_info_table {
	my ($info_table) = @_;
	
	my $res = "";
	
	# Print the local system data.
	$res .= '<table class="info">'."\n";
	if ( $info_table->{TITLE} ) { # Table title
		$res .= '<tr><th colspan="2" class="title">'.$info_table->{TITLE}.'</th></tr>'."\n";
	}
	foreach my $row ( @{$info_table->{ROWS}} ) {
		my $first_column = $row->{LEFT};
		my $second_column = $row->{RIGHT};
		my $background_color = $row->{BG_COLOR};
		my $foreground_color = $row->{FG_COLOR};
		$res .= info_line( $first_column, $second_column, $background_color, $foreground_color );
	}
	$res .= '</table>';
	
	return $res;
}
#----------------------------------------------------------------------

# Prints a single line of a two-column table.
#
# Arguments:
# $1 - The first column.
# $2 - The second column.
# $3 (optional) - The foreground color of the second column.
# $4 (optional) - The background color of the second column.
sub info_line {
	my ($first_column, $second_column, $background_color, $foreground_color ) = @_;
	
	if ( !defined $second_column ) {
		$second_column = '<i style="color:'.$neutral_gray.'">(undef)</i>';
	}
	
	# Compile the style of the first column
	my $style1 = "white-space:nowrap;";
	
	$style1 = ' style="'.$style1.'"';
	
	# Compile the style of the second column
	my $style2 = "";
	
	$style2 .= "text-align:right;";
	
	if ( $foreground_color ) {
		$style2 .= "color:$foreground_color;";
	}
	if ($background_color) {
		$style2 .= "background-color:$background_color;";
	}
	
	$style2 = ' style="'.$style2.'"';
	
	# Print the line with appropriate decorations.
	my $res = "<tr><th".$style1.">".$first_column."</th>\n"
			."<td".$style2.">".$second_column."</td>\n"
			."</tr>\n";
	
	return $res;
}
#----------------------------------------------------------------------

sub JournalReport {
	my ($jstruct) = @_;
	
	my $no_problems = 1;
	
	my $pt = "";
	
	$pt .= problem_table_header();
	
	my $id = 0;
	my $journal_id = extract_filename($jstruct->{FILE});
	
	foreach my $obj ( @{$jstruct->{ARR}} ) {
		next if ref($obj) eq "";
		
		if ( ref($obj) ne "HASH" ) {
			warning "Wrong jstruct element: '$obj'";
			next;
		}
		
		$no_problems = 0 if $no_problems;
		
		my $box_id = $journal_id.".".(++$id);
		$box_id =~ s/([^\w\-_])/./g;
		
		$pt .= problem_table_line( $obj, $jstruct, $box_id );
	}
	
	$pt .= "</table>\n";

	my $res = "";
	
	if ( defined $jstruct->{INFO_TABLES} ) {
		$res .= print_info_table($_) foreach ( @{$jstruct->{INFO_TABLES}} );
	}

	if ( $no_problems ) {
		$res .= '<h4>Problem summary:</h4>';
		$res .= '<p style="font-weight:bold;color:'.$dark_green.'">'
				.'No problems were detected.</p>';
		return $res;
	}
	else {
		$res .= $pt;
	}
	
	return $res;
}
#----------------------------------------------------------------------

# Prints a header of the problem table.
sub problem_table_header {
	my $res = <<HERETEXT;
<h4>Problem Summary</h4>

<p>Click on lines in the table to see the details about each problem.</p>

<table class="problem_table">
<thead><tr style="background-color:$gray">
<th>Test Name</th><th width="17%">Severity</th><th width="23%">Resolution</th>
</tr></thead>

HERETEXT
}

#---------------------------------------------------------------------

sub problem_table_line {
	my ( $data, $jstruct, $box_id ) = @_;
	
	my $severity = ($data->{SEVERITY} or "");
	my $gen_severity = ( $severity_map{$severity} or $severity );
	my $severity_html = html($severity);
	$severity_html =~ s/ +/&nbsp;/g; # Make spaces nonbreakable
	
	my $color = "red";
	if ( $severity =~ /(quelled|waived)/i ) {
		# Anything what is quelled is harmless, so use neutral gray
		# for all quelled problems.
		$color = "gray";
	}
	elsif ( $gen_severity eq "failed" || $gen_severity eq "testsuite_error" ) {
		$color = "red"; # Use fluorescent red for various failures.
	}
	elsif ( $gen_severity eq "warning" ) {
		$color = "yellow"; # Use fluorescent yellow for warnings.
	}
	elsif ( $gen_severity eq "fip" ) {
		$color = "yellow"; # Use fluorescent yellow for FIPs.
		
		$severity_html .= '<span class="view_srv">&nbsp;&nbsp;'
			.'<a href="tests_help.pl#FIP" title="Help" onclick="event.cancelBubble=true" target="_blank">' # HELP icon
				.'<img src="images/question.png" width="16" height="17" alt="[?]" />'
			.'</a></span>';
	}
	elsif ( $gen_severity eq "notice" || $gen_severity eq "untested" ) {
		# Actually, untested test cases shouldn't displayed
		$color = "gray"; # Use neutral gray for harmless messages.
	}
	else {
		warning "Unrecognized severity: '$severity'";
		$severity_html = "Unrecognized severity: '$severity_html'";
		$severity = $gen_severity = $data->{SEVERITY} = 'testsuite_error';
	}

	my $test_case  = ( $data->{TEST_NAME} or "..." );
	if ( defined $data->{TEST_POINT} ) {
		$test_case .= ' '.$data->{TEST_POINT};
	}
	
	my $res = "";
	
	# Give the box a unique ID.
	my $message_box_id = $box_id;
	
	# Start the line
	# (with attached script that opens the detail block on clicking).
	$res .= '<tr id="tr_'.$message_box_id.'" class="ptl_'.$color.'" title="Click here!"'
			.' onclick="flip(\''.$message_box_id.'\');">';
	
	# Print the test name and test point (if available).
	$res .= "<td>";
	if ( defined $data->{TEST_NAME} ) {
		$res .= html( $data->{TEST_NAME} );
	} else {
		$res .= "<i>(none)</i>";
	}
	
	if ( defined $data->{TEST_POINT} ) {
		$res .= ' <span style="color:'.$dark_blue.'">'.html($data->{TEST_POINT}).'</span>';
	}
	$res .= "</td>\n";

	# Print severity
	$res .= '<td class="fw1">'.$severity_html."</td>\n";
	
	# Print resolution
	$res .= '<td class="fw2" id="td2_'.$message_box_id.'">'.html($data->{RESOLUTION} or "")."</td>\n";

	# Finish the line.
	$res .= "</tr>\n";
	
	# Print the detail block, which is initially hidden and unrolls when
	# the table row is clicked (via the javascript).
	$res .= '<tr>'."\n";
	$res .= '<td colspan="3" class="pdu">';
	$res .= '<div class="ptd_'.$color.'" id="'.$message_box_id.'"'
				.' style="display:none">';
	
	# Test description (SimpleJournal)
	if ( defined $data->{DESCRIPTION} && $data->{DESCRIPTION} ne "" ) {
		$res .= '<h4>Test Description:</h4>'."\n";
		$res .= '<pre style="overflow:auto">'.html_excl($data->{DESCRIPTION}).'</pre>'."\n";
		$res .= '<hr />';
	}
	
	# Message from test journal.
	if ( defined $data->{MESSAGE} && $data->{MESSAGE} ne "" ) {
		$res .= '<h4>Message from the test:</h4>'."\n";
		$res .= '<pre style="overflow:auto">'.html_excl($data->{MESSAGE}).'</pre>'."\n";
		$res .= '<hr />';
	}
	
	# Test comment (TET journal, SimpleJournal)
	if ( defined $data->{COMMENT} && $data->{COMMENT} ne "" ) {
		$res .= '<h4>Comment:</h4>'."\n";
		$res .= '<pre style="overflow:auto">'.html_excl($data->{COMMENT}).'</pre>'."\n";
	}
	
	if ( $gen_severity eq 'fip' ) {
		my $fip_test_case = $test_case;
		$fip_test_case =~ s/'/_/g;
		my $fip_comments = ""; # TODO:

		# FIP form
		$res .= '<div class="view_srv">';
		$res .= '<h4>Please resolve this test case manually:</h4>'."\n";
		$res .= '<form id="fip_'.$message_box_id.'_form" method="post" action="">'."\n";
			$res .= '<input type="hidden" name="testcase" value="'.html($fip_test_case).'"/>';
			$res .= '<span id="fip_'.$message_box_id.'_passfail" class="ptl_yellow" style="white-space: nowrap;padding:5pt;">';
				$res .= '<label><input type="radio" name="result" id="fip_'.$message_box_id.'_pass" value="pass" onclick="fip_proceed(\''.$message_box_id.'\', 0)"/>Pass</label> &nbsp; ';
				$res .= '<label><input type="radio" name="result" id="fip_'.$message_box_id.'_fail" value="fail" onclick="fip_proceed(\''.$message_box_id.'\', 0)"/>Fail</label>';
			$res .= '</span>'."\n";
			$res .= '<table>';
			$res .= '<tr><td style="valign:center">';
			$res .= '<font size="-1"><b>Additional comments:</b></font>';
			$res .= '</td><td>';
			$res .= '<textarea name="comment" id="fip_'.$message_box_id.'_comment" rows="3" cols="60" onblur="fip_proceed(\''.$message_box_id.'\', 0)">'.$fip_comments.'</textarea>';
			$res .= '</td></tr>';
			$res .= '</table>';
			$res .= '<table width="100%">';
			$res .= '<tr><td align="left">';
			$res .= '<input type="submit" name="fip_'.$message_box_id.'_next" value="&#9660; Next FIP case &#9660;" onclick="return fip_proceed(\''.$message_box_id.'\', \'next\')"/>';
			$res .= '</td><td align="right">';
			$res .= '<input type="submit" name="fip_'.$message_box_id.'_prev" value="&#9650; Previous FIP case &#9650;" onclick="return fip_proceed(\''.$message_box_id.'\', \'prev\')"/>';
			$res .= '</td></tr>';
			$res .= '</table>';
		$res .= '</form>';
		$res .= '<hr />';
		$res .= '</div>';
		$res .= '<script language="javascript" type="text/javascript">if (typeof(fip_ids) != "undefined") { fip_ids[fip_tmpnum] = \''.$message_box_id.'\'; fips_n_by_id[\''.$message_box_id.'\'] = fip_tmpnum; fips_n_by_testcase[\''.$fip_test_case.'\'] = fip_tmpnum; fip_resolved[fip_tmpnum] = 0; fip_tmpnum++; }</script>';

		# For html standalone report:
		$res .= '<div style="display:none" id="fip_'.$message_box_id.'_cmntarea">'
			.'<h4>Additional Comments:</h4>'
			.'<pre id="fip_'.$message_box_id.'_cmntpre"></pre>'
			.'</div>';
	}

	# Link to the point in the journal.
	{
		$res .= '<h4>Open the journal:</h4>'."\n";
		$res .= '<p>';
		
		my $href = $jstruct->{FILE};
		$href .= '?line='.$data->{LINE_NUM}.'#line'.$data->{LINE_NUM} if defined $data->{LINE_NUM};
		
		my $text = $jstruct->{FILE};
		$text .= ':'.$data->{LINE_NUM} if defined $data->{LINE_NUM};
		
		$res .= local_href($href, $text);
		
		$res .= '</p>'."\n";
		$res .= '<hr />';
	}
	
	# Problem info from problem DB
	my $pr_info = $data->{PROBLEM_INFO};
	if ( $pr_info ) {
		$res .= '<h4>Problem info:</h4>'."\n";
		
		if ( $pr_info->{'comment'} ) {
			# Contains html text
			$res .= '<p>'.$pr_info->{'comment'}.'</p>'."\n";
		}
		if ( $pr_info->{'lsb_report_id'} ) {
			my $pr_id = $pr_info->{'lsb_report_id'};
			my $link = "https://www.linuxfoundation.org/lsbprs/pr/".$pr_id;
			$res .= '<p>Reported to the LSB Committee as'
					.' <a href="'.$link.'" target="_blank">PR '.$pr_id.'</a>.</p>'."\n";
		}
		
		$res .= '<hr />';
	}
	
	# "Report bug in test" button.
	if (  $color eq "red"    # If a bug
		  && (!$pr_info || !$pr_info->{'lsb_report_id'} ) # And not yet reported
	   ) 
	{
		local *quot = sub {
			my ($s) = @_;
			$s = html($s);
			$s =~ s/'/\\'/g;
			return $s;
		};
		$test_case = quot($test_case);

		my $test_suite = quot($jstruct->{TESTSUITE} or "undef");
		
		my $file = quot($jstruct->{FILE} or "");
		
		$res .= <<HERETEXT;
<div align="right">
	<input type="button" value="Report Bug in the Test" 
		onclick="javascript:bug_report('$test_suite', '$test_case', '$severity', '$file');"
		title="Press this button if you believe the problem is in the test, not in your system." />
</div>
HERETEXT
		#
	}
	
	# Finish the table line
	$res .= "</div>";
	$res .= "</td></tr>\n";

	return $res;
}
#----------------------------------------------------------------------

sub Compile {
	# Read the runconfig file
	my $runconfig = read_config( $result_dir."/runconfig" );
	is_ok($runconfig) or return error "Failed to read runconfig", $Error::Last;
	
	my $tests = $runconfig->{SECTIONS};
	
	# Information in the 'runconfig' file is for non-executed tests
	# For executed tests information is obtained from <test>.info files.
	
	# Read tests' .info files
	my @info_files = glob($result_dir."/results/*.info");
	foreach my $info_file ( @info_files ) {
		my $test_info = read_config( $info_file );
		my $test_name = $test_info->{NAME} or next;
		$tests->{$test_name} = $test_info;
	}
	
	for my $test_name ( keys %$tests ) {
		my $test_info = $tests->{$test_name};
		
		if ( !$test_info->{VERDICT} ) { $test_info->{VERDICT} = "incomplete"; }
		
		$test_info->{REPORT} = "";
		
		# Title:
		$test_info->{REPORT} .= '<h2>';
		
		$test_info->{REPORT} .= '<a name="'.$test_info->{NAME}.'"></a>'
				.'<font color="#336699">'.html($test_info->{DISPLAYNAME}).'</font>';
		
		if ( defined $test_info->{VERSION} ) {
			$test_info->{REPORT} .= ' &nbsp; <font size="-1" color="'.$neutral_gray.'">'
						.'v. '.html($test_info->{VERSION}).'</font>';
		}
		if ( defined $test_info->{MANUAL} ) {
			$test_info->{REPORT} .= ' &nbsp; <font size="-1">(Manual test)</font>';
		}
		
		if ( $test_info->{TRIAL_USE} && $cert_mode ) {
			$test_info->{REPORT} .= ' &nbsp; <span style="color:'.$fluorescent_green.'">'
					.'TRIAL USE</span>';
		}
		
		$test_info->{REPORT} .= '</h2>'."\n";
		
		# Read .report.part1 file
		my $report_filename = $result_dir."/results/".$test_info->{NAME}.".report.part1";
		my $report_part = read_file( $report_filename );
		if ( is_ok($report_part) ) {
			$test_info->{REPORT} .= $report_part;
		}
		
		# Test info table
		if ( !$test_info->{STARTED} ) {
			if ( $test_info->{MANUAL} ) {
				$test_info->{REPORT} .= "<p>The test hasn't been prepared.</p>\n";
			} else {
				$test_info->{REPORT} .= "<p>The test hasn't been executed.</p>\n";
			}
		} else {
			$test_info->{REPORT} .= print_testinfo_table( $test_info );
		}
		
		# Add .report.part2 file
		$report_filename = $result_dir."/results/".$test_info->{NAME}.".report.part2";
		$report_part = read_file( $report_filename );
		if ( is_ok($report_part) ) {
			$test_info->{REPORT} .= $report_part;
		}
		
		$test_info->{REPORT} .= "\n<hr />\n\n";
	}
	
	my @auto_tests = grep {!$_->{MANUAL} && !($cert_mode && $_->{TRIAL_USE})} values %$tests;
	my $verdict = undef;
	if ( @auto_tests ) {
		$verdict = overall_verdict( map {$_->{VERDICT}} @auto_tests );
	}
	my @man_tests = grep {$_->{MANUAL} && !($cert_mode && $_->{TRIAL_USE})} values %$tests;
	my $verdict_man = undef;
	if ( @man_tests ) {
		$verdict_man = overall_verdict( map {$_->{VERDICT}} @man_tests );
	}
	
	my $res = "";
	
	$res .= Head( $runconfig );
	
	my @sort_order = ();
	$res .= Tree_toc( $tests, \@sort_order );

	$res .= '<div id="fip_header" style="display:none;border:1px solid black; background-color:#FFFF90;margin:10pt;padding:2pt;"></div>'."\n";
	$res .= "\n<hr />\n\n";

	my @tests = sort  { in_array($a, @sort_order) <=> in_array($b, @sort_order) }  keys %$tests;
	
	foreach my $test_name ( @tests ) {
		if ( !defined $tests->{$test_name}->{REPORT} ) {
			warning "No report for '$test_name'";
			next;
		}
		$res .= $tests->{$test_name}{REPORT};
	}
	
	$res .= Footer();
	
	# Write the report to file
	my $report_file = $result_dir."/report.htm";
	is_ok write_string_as_file( $report_file, $res)
		or complain "Failed to write the report to '$report_file'", $Error::Last;
	
	return $verdict, $verdict_man;
} # End of Compile()
#-----------------------------------------------------------------------

sub print_testinfo_table {
	my ($test_info) = @_;
	
	my $res = "";
	
	# Start/stop time
	my $start_time = $test_info->{START_TIME};
	my $finish_time = $test_info->{FINISH_TIME};
	
	$res .= '<table class="info">'."\n";
	
	if ( $start_time ) {
		$res .= info_line('Started At', format_time($start_time) );
		if ( $finish_time ) {
			$res .= info_line('Execution Time', format_duration($finish_time - $start_time) );
		}
	} else {
		if ( !$test_info->{MANUAL} ) {
			$res .= info_line('Started At', "NOT STARTED", $fluorescent_red );
		}
	}
	
	$res .= info_line( 'Verdict:', 
			verdict_text( $test_info->{VERDICT} ),
			verdict_color( $test_info->{VERDICT} )
		);
	
	$res .= '</table>'."\n";
	
	return $res;
}

sub verdict {
	my ($stats) = @_;
	
	# Deduce the verdict
	my $verdict = undef;
	if ( $stats->{FAILURES} || $stats->{TESTSUITE_ERRORS} ) {
		$verdict = "failed";
	}
	elsif ( $stats->{WARNINGS} ) {
		$verdict = "warning";
	}
	elsif ( defined $stats->{TESTS_EXPECTED} && defined $stats->{TESTS_TOTAL}
		  && $stats->{TESTS_EXPECTED} != $stats->{TESTS_TOTAL} ) 
	{
		$verdict = "warning";
	}
	else {
		$verdict = "passed";
	}
	return $verdict;
}
#-----------------------------------------------------------------------

my %verdict_sort = ();
$verdict_sort{$_} = scalar keys %verdict_sort
	for ('run_problems', 'failed', 'undefined', 'incomplete', 'warning', 'passed', 'skipped', '');

sub overall_verdict {
	my $res = "";
	
	foreach my $verdict ( @_ ) {
		if ( !$verdict ) { $verdict = "incomplete"; }
		if ( $verdict_sort{$verdict} < $verdict_sort{$res} ) {
			$res = $verdict;
		}
	}
	
	if ( $res eq 'undefined' ) { warning "Undefined verdict"; }
	
	return $res;
}
#----------------------------------------------------------------------

sub Head {
	my ($runconfig) = @_;
	
	my $res = "";
	
	# TODO: read from file ../webui/public_html/css/report.css

	$res .= <<HEREDOC;
<html>
<head>
	<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
	<title>Test Results</title>

<style type="text/css">

html, body {
	font-family: sans-serif;
	font-size: x-small;
}

h1 { border-bottom: 1px solid  #aaa; }
h2 { border-bottom: 1px dashed #aaa; }

a {
	text-decoration: none;
	color: #336699;
	background: none;
}
a:hover {
	text-decoration: underline;
	color: #336699;
}

.toclink {
	font-weight: bold;
}

pre {
    padding: 1em;
    border: 1px dashed #2f6fab;
    color: black;
    background-color: #f9f9f9;
    line-height: 1.1em;
    overflow: auto;
}

.errmsg {
	background: #f9ffaa;
	border: 1px dashed red;
	color: red;
	margin-bottom: 15px;
	padding: 3pt 15pt;
	font-size: 120%;
}

table.info {
	margin: 1em 1em 6pt 0;
	background: #f9f9f9;
	border: 1px #aaaaaa solid;
	border-collapse: collapse;
}
table.info th, table.info td {
	border: 1px #aaaaaa solid;
	padding: 0.2em;
	padding-left: 0.4em;
	padding-right: 0.4em;
}

table.toc {
	border-collapse: collapse;
	background: #E6E6E6;
}
table.toc th, table.toc td {
	border: 1px #FFFFFF solid;
	padding: 0.2em;
	padding-left: 0.4em;
	padding-right: 0.4em;
}

table.info th, table.toc th {
	background: #f2f2f2;
	text-align: left;
}
table.info th.title {
	background: #E5E5E5;
}

table.problem_table {
	width: 100%;
	border-collapse: collapse;
	table-layout: fixed;
}
table.problem_table thead th {
	background-color: $gray;
}

table.problem_table tr td {
	border: 1px solid white;
}

td.pd, td.pdu { /* Problem details */
	margin: 0;
	padding: 0;
	border: 0;
}
td.pdu {
	width:100%;
}

.fw1 { width:12%; text-align: center; }
.fw2 { width:25% }

.ptl_gray       { background: #E6E6E6;}
.ptl_gray:hover { background: #D6D6F0;}

.ptl_green       { background: #CDFFCD;}
.ptl_green:hover { background: #A0FFA0;}

.ptl_yellow       { background: #FFFFCD;}
.ptl_yellow:hover { background: #FFFF90;}

.ptl_red   { background: #FFCDCD;}
.ptl_red:hover { background: #FFA0A0;}

.ptl_gray, .ptl_green, .ptl_yellow, .ptl_red { cursor: pointer }

.ptd_gray, .ptd_green, .ptd_yellow, .ptd_red { 
	padding: 1pt 16pt 5pt 16pt;
	margin: 3pt 5pt 10pt 5pt;
}
.ptd_gray   { border-left: 5pt solid #D6D6F0 }
.ptd_green  { border-left: 5pt solid #A0FFA0 }
.ptd_yellow { border-left: 5pt solid #FFFF90 }
.ptd_red    { border-left: 5pt solid #FFA0A0 }

.view_srv { display: none; }

.legal-text { width: 100%; font-size: xx-small; text-align:center; color: gray; }

</style>
</head>
HEREDOC

	my $rep_platform = "All";
	if ( $runconfig->{'HOST_architecture'} ) {
		$rep_platform = $runconfig->{'HOST_architecture'};
		if ( $rep_platform eq 'x86' ) { $rep_platform = 'IA32'; }
		if ( $rep_platform eq 'x86-64' ) { $rep_platform = 'AMD64'; }
		if ( !in_array($rep_platform, "IA32", "AMD64", "IA64", "PPC32", "PPC64", "S390", "S390X") ) {
			$rep_platform = "All";
		}
	}
	my $op_sys = "All";
	if ( $runconfig->{'HOST_OS'} ) {
		for my $s ( "Asianux", "BOSS", "Booyo", "Debian", "Fedora", "Linpus", "Mandriva",
				"OpenSuse", "Red Flag", "Red Hat", "Sun Wah", "SuSE", "ThizLinux", "Turbolinux",
				"Ubuntu", "United Linux", "Xandros" )
		{
			my $re = $s;
			$re =~ s/\s+/.?/;
			if ( $runconfig->{'HOST_OS'} =~ /$re/ ) {
				$op_sys = $s;
				last;
			}
		}
	}

	$res .= <<HEREDOC;
<body>
<!-- Bugreport form. Used when one clicks "Report bug" button. -->
<form name="bugreport" style="display:none" method="get" action="http://bugs.linuxbase.org/enter_bug.cgi" target="_blank">
<input type="hidden" name="product" value="Distribution Tests" />
<input type="hidden" name="format" value="problemreport" />
<input type="hidden" name="rep_platform" value="$rep_platform" />
<input type="hidden" name="op_sys" value="$op_sys" />
<input type="hidden" name="short_desc" value="" />
<input type="hidden" name="comment" value="" />
</form>

<script language="javascript" type="text/javascript">
// <![CDATA[
function flip (message_box_id) {
	msgbox = document.getElementById(message_box_id);
	if (msgbox.style.display != 'none') {
		msgbox.style.display = 'none';
	}
	else {
		msgbox.style.display = '';
	}
}
function bug_report(test_suite, test_case, severity, file) {
	document.bugreport.short_desc.value = test_suite+": "+test_case+" "+severity;
	document.bugreport.comment.value = "Enter a description of the problem here\\n\\n---\\nDistribution Checker: Please use the 'Add an attachment' button below to attach the following test journal:\\n "+file;
	document.bugreport.submit();
}

var fip_tmpnum = 0;
var fip_ids = new Array();
var fips_n_by_id = new Array();
var fips_n_by_testcase = new Array();
var fip_resolved = new Array();

function fip_go_to(fip_n) {
	if ( fip_n < 0 ) {
		// Go to the header
		fip_header = document.getElementById('fip_header');
		if ( fip_header ) {
			fip_header.scrollIntoView(true);
		} else {
			window.alert('Error: no element: fip_header');
		}
		return;
	}
	
	var box_id = fip_ids[fip_n];

	// unhide row details
	msgbox = document.getElementById(box_id);
	if ( msgbox ) {
		msgbox.style.display = '';
	} else {
		window.alert('Error: no element: ' + box_id);
	}
	// scroll to row
	tr_elem = document.getElementById('tr_'+box_id);
	if ( tr_elem ) {
		tr_elem.scrollIntoView(true);
	} else {
		window.alert('Error: no element: tr_'+box_id);
	}
}

function fip_repack(self) {
	self.disabled = 'true';
	if ( typeof(ajax_call_get) != 'undefined' ) {
		var prev_timeout = timeout_show_progress; // Do a workaround to set less timeout
		timeout_show_progress = 250;
		ajax_call_get("action=repack&details="+webui_details);
		timeout_show_progress = prev_timeout;
	}
}

function fip_update_header() {
	var fip_header = document.getElementById('fip_header');
	if ( !fip_header ) return;
	
	var fips_num = fip_ids.length;
	if ( !fips_num ) return;

	fip_header.style.display='';

	var fips_unresolved = 0;
	var fip_first_unresolved_n = -1;
	for (var i=0; i<fip_ids.length; ++i) {
		if ( !fip_resolved[i] ) {
			fips_unresolved++;
			if ( fip_first_unresolved_n < 0 ) {
				fip_first_unresolved_n = i;
			}
		}
	}

	if ( fips_unresolved > 0 ) {
		var fip_header_text = 'There '+(fips_num==1?'is <b>one</b>':'are <b>'+fips_num+'</b>')+' test case'+(fips_num==1?'':'s')+' labeled <b>"FIP"</b>, which means "Further Information Provided". Such test cases require manual resolution.<br />';

		if ( fips_unresolved == 1 ) {
			if ( fips_unresolved != fips_num )
				fip_header_text += ' There is <b>one unresolved</b> FIP case left.';
			fip_header_text += ' Please, choose "pass" or "fail" for this test case.';
		} else {
			if ( fips_unresolved != fips_num )
				fip_header_text += ' There are <b>'+fips_unresolved+' unresolved</b> FIP cases left.';
			fip_header_text += ' Please, go through these test cases and choose "pass" or "fail" for each one.';
		}

		fip_header_text += '<br/> &nbsp; <input type="button" value="&#9660; Go to the first unresolved &quot;FIP&quot; case &#9660;" onclick="javascript:fip_go_to('+fip_first_unresolved_n+');"/>';
		
		fip_header.innerHTML = fip_header_text;
	}
	else {
		fip_header.style.backgroundColor = '#A0FFA0';
		var fip_header_text = 'There '+(fips_num==1?'is one':'are '+fips_num)+' test case'+(fips_num==1?'':'s')+' labeled "FIP", which means "Further Information Provided".<br /> &nbsp; <b>All the "FIP" cases were resolved.</b>';

		if ( fip_need_repack ) {
			fip_header_text += '<br/><br/> &nbsp; <b>Please push this button to pack the latest information on FIP cases. It may take a few seconds.</b>';
			
			fip_header_text += '<br/> &nbsp; <input type="button" value="&#9658; Repack &#9668;" onclick="javascript:fip_repack(this);"/>';
		}
		else {
			fip_header_text += '<br/> &nbsp; <b>The results package: <a href="get.pl'+results_tarball+'">'+results_tarball+'</a></b>.';
			fip_header_text += '<br/> &nbsp; <input type="button" value="&#9660; Go to the first &quot;FIP&quot; case &#9660;" onclick="javascript:fip_go_to(0);"/>';
		}
		
		// TODO: finish

		fip_header.innerHTML = fip_header_text;
	}
}

function fip_update(box_id) {
	var elem_pass = document.getElementById('fip_'+box_id+'_pass');
	var elem_fail = document.getElementById('fip_'+box_id+'_fail');
	if (!elem_pass || !elem_fail) {
		window.alert('Error: no element: fip_'+box_id+'_pass or fip_'+box_id+'_fail');
		return false;
	}
	var color = '';
	var resolution = '';
	if ( elem_pass.checked ) {
		color = 'green';
		resolution = 'pass';
	}
	else if ( elem_fail.checked ) {
		color = 'red';
		resolution = 'fail';
	}
	
	msgbox = document.getElementById(box_id);
	if ( msgbox && color ) {
		msgbox.className='ptd_'+color;
	}
	tr_elem = document.getElementById('tr_'+box_id);
	if ( tr_elem && color) {
		tr_elem.className='ptl_'+color;
	}
	td2_elem = document.getElementById('td2_'+box_id);
	if ( td2_elem && resolution ) {
		td2_elem.innerHTML=resolution;
	}
	passfail_elem = document.getElementById('fip_'+box_id+'_passfail');
	if ( passfail_elem && color ) {
		passfail_elem.className='ptl_'+color;
	}
	return true;
}

function fip_proceed(box_id, next_or_prev) {
	if ( !fip_update(box_id) ) return false;

	var cur_fip_n = fips_n_by_id[box_id];
	var next_fip_n = -1;
	if ( next_or_prev == 'next' ) {
		next_fip_n = cur_fip_n+1;
		if ( next_fip_n >= fip_ids.length )
			next_fip_n = -1;
	}
	else if ( next_or_prev == 'prev' ) {
		next_fip_n = cur_fip_n-1;
	}
	// next_fip_n may be < 0, which means "return to the header".

	if ( typeof(ajax_call_post) != 'undefined' ) {
		ajax_call_post('action=fip&details='+webui_details+(next_or_prev ? '&next='+next_fip_n : ''), 'fip_'+box_id+'_form');
	}
	fip_resolved[cur_fip_n] = 1;
	fip_update_header();

	if ( next_or_prev )
		setTimeout('fip_go_to('+next_fip_n+')', 200);

	return false; // cancel form submit
}

// ]]>
</script>

HEREDOC
	
	# Main title
	$res .= '<h1>Automatic test results for '.html($runconfig->{'HOST_OS'})
			.' on '.html($runconfig->{'HOST_architecture'}).'</h1>';
	
	if ( -f "$result_dir/log" )  {
		$res .= '<p>The complete console output log is available '
					.local_href( "log", 'here' ).'.</p>';
	}
	if ( -f "$result_dir/verbose_log" )  {
		$res .= '<p>The '.local_href( "verbose_log", 'verbose log' )
					.' is also available.</p>';
	}
	
	$res .= '<table class="info">';
	$res .= info_line("Standard", $runconfig->{'STANDARD'});
	$res .= info_line("Machine", $runconfig->{'HOST_machine'}." (".$runconfig->{'HOST_architecture'}.")");
	$res .= info_line("Package Manager",  $runconfig->{'HOST_package_manager'});
	$res .= info_line("Hostname",  $runconfig->{'HOST_hostname'});
	$res .= info_line("Kernel",  $runconfig->{'HOST_kernel'});
	$res .= info_line("Operation System",  $runconfig->{'HOST_OS'});
	$res .= info_line("/etc/issue",  $runconfig->{'HOST_etc_issue'});
	$res .= '</table>';
	
	return $res;
}

sub Footer {
	my $res = "";
	
	$res .= <<HEREDOC;
<script type="text/javascript" src="fips.js"></script>

<script language="javascript" type="text/javascript">
// <![CDATA[

function fip_update_form() {
	if ( typeof(fip_results) == 'undefined' ) return;
	if ( typeof(fip_comments) == 'undefined' ) return;
	
	for ( var testcase in fip_results ) {
		var fip_n = fips_n_by_testcase[testcase];
		var box_id = fip_ids[fip_n];
		if ( fip_results[testcase] ) {
			var pass_fail_id = 'fip_'+box_id+'_'+fip_results[testcase];
			var elem_pass_fail = document.getElementById(pass_fail_id);
			if ( elem_pass_fail ) {
				elem_pass_fail.checked = 'checked';
				fip_resolved[fip_n] = 1;
				fip_update(box_id, 0);
			}
		}
		if ( fip_comments[testcase] ) {
			var elem_comment = document.getElementById('fip_'+box_id+'_comment');
			if ( elem_comment ) {
				elem_comment.text = fip_comments[testcase];
			}
			var elem_cmntarea = document.getElementById('fip_'+box_id+'_cmntarea');
			var elem_cmntpre = document.getElementById('fip_'+box_id+'_cmntpre');
			if ( elem_cmntarea && elem_cmntpre ) {
				elem_cmntpre.text = fip_comments[testcase];
				elem_cmntarea.style.display = '';
			}
		}
	}
}
fip_update_form();

// ]]>
</script>

</body>
</html>
HEREDOC

	return $res;
}
#----------------------------------------------------------------------
my @toc_groups = ("Certification Tests", "Unofficial Tests",
	"Command Tests", "Static Interface Tests", "Runtime Interface Tests", "Moblin Specific Tests", "Automated Application Battery Tests", "Manual Tests" );

sub Tree_toc {
	my ($tests, $sort_arr) = @_;
	
	my $res = "";
	
	# Building a tree structure;
	my $tree = {};
	
	foreach my $test_info ( values %$tests ) {
		my $tmp = $tree;
		
		$test_info->{TREE} = "Undefined" if !defined $test_info->{TREE};
		
		# Enter the groups
		foreach my $group_name ( split "/", $test_info->{TREE} ) {
			if ( $group_name =~ s/^<([^>]*)>/$1/ ) {
				# Standard group
				( in_array($group_name, @toc_groups) )
					or warning "Wrong group name: '$group_name'";
			}
			$tmp->{CHILDREN}->{$group_name} ||= {};
			$tmp = $tmp->{CHILDREN}->{$group_name};
		}
		# If this test has subtests, we shouldn't lose them.
		if ( defined $tmp->{CHILDREN}->{$test_info->{DISPLAYNAME}}->{CHILDREN} ) {
			$test_info->{CHILDREN} = $tmp->{CHILDREN}->{$test_info->{DISPLAYNAME}}->{CHILDREN};
		}
		# Add the test
		$tmp->{CHILDREN}->{$test_info->{DISPLAYNAME}} = $test_info;
	}
	
	# Print the tree
	$res .= '<table class="toc">'."\n";
	
	if ( $tree->{CHILDREN} ) {
		$res .= print_toc_subtree( $tree, 0, $sort_arr );
	}
	
	$res .= '</table>'."\n";
	
	return $res;
}

sub print_toc_subtree {
	my ($subtree, $deep, $sort_arr) = @_;
	
	my $res = "";

	if ( $subtree->{CHILDREN} ) {
		my @keys = keys %{$subtree->{CHILDREN}};
		
		@keys = sort {
				my $ap = in_array($a, @toc_groups);
				my $bp = in_array($b, @toc_groups);
				
				return $ap <=> $bp if ( $ap && $bp );
				return lc($a) cmp lc($b) if ( !$ap && !$bp );
				return 1 if !$ap;
				return -1 if !$bp;
			} @keys;
		
		foreach my $node ( @keys ) {
			my $child = $subtree->{CHILDREN}->{$node};
			push @$sort_arr, $child->{NAME} if $child->{NAME};
			$res .= print_toc_node( $node, $child, $deep, $sort_arr );
		}
	}
	
	return $res;
}

sub print_toc_node {
	my ($name, $subtree, $deep, $sort_arr) = @_;
	
	my $res = "";
	
	my $padding = 5 + $deep * 6;
	my $style_padding = "padding-left:$padding pt";
	
	my $color = $deep <= 3 ? 160 + 30 * $deep : 250;
	$color = sprintf ('#%02X%02X%02X', $color, $color, 250);
	my $style_color = "background:$color";
	
	if ( $subtree->{NAME} ) {
		my $test_info = $subtree;
		
		# Test name as a link to the corresponding part of the report.
		$res .= '<tr style="'.$style_color.'">';
		$res .= '<td style="'.$style_padding.'"'
				.(defined $test_info->{VERSION} ? '' : ' colspan="2"').'>'."\n";
		$res .= '<a href="#'.$test_info->{NAME}.'" class="toclink">'
					.html($test_info->{DISPLAYNAME}).'</a>';
		$res .= '</td>';
		
		# Version
		if ( defined $test_info->{VERSION} ) {
			$res .= '<td style="color:'.$dark_gray.'">'
					.'v '.html($test_info->{VERSION})
					.'</td>';
		}
		
		$res .= '<td style="background-color:'.verdict_color($test_info->{VERDICT}, $color).'">'
				.verdict_text($test_info->{VERDICT}, "Undefined").'</td>'."\n";
		
		$res .= '</tr>';
	}
	else {
		$res .= '<tr><th style="'.$style_color.";".$style_padding.'" colspan="3">'."\n";
		$res .= html($name);
		$res .= '</th></tr>';
	}
	if ( $subtree->{CHILDREN} ) {
		$res .= print_toc_subtree( $subtree, $deep + 1, $sort_arr );
	}
	
	return $res;
}
#----------------------------------------------------------------------

sub verdict_color {
	my ($verdict, $default_color) = @_;
	
	if ( !$verdict || $verdict eq "undefined" || $verdict eq "skipped" ) { # undef
		return $default_color;
	}
	elsif ( $verdict eq "run_problems" ) {
		return $fluorescent_red;
	}
	elsif ( $verdict eq "passed" ) {
		return  $fluorescent_green;
	}
	elsif ( $verdict eq "warning" || $verdict eq "incomplete" ) {
		return $fluorescent_yellow;
	}
	elsif ( $verdict eq "failed" ) {
		return $fluorescent_red;
	}
	else {
		warning "Unhandled verdict: '$verdict'";
	}
}

sub verdict_text {
	my ($verdict, $default_text) = @_;
	$default_text = "" if !defined $default_text;
	
	if ( !$verdict || $verdict eq "undefined" ) {
		return $default_text;
	}
	elsif ( $verdict eq "run_problems" ) {
		return "There are problems";
	}
	elsif ( $verdict =~ m/^(passed|warning|incomplete|skipped|failed)$/ ) {
		return ucfirst $verdict;
	}
	else {
		warning "Unhandled verdict: '$verdict'";
	}
}

sub print_stats {
	my ($stats) = @_;
	
	my $tests_total  = $stats->{TESTS_TOTAL};
	my $tests_passed = $stats->{PASSED};
	my $failures = $stats->{FAILURES};
	my $warnings = $stats->{WARNINGS};
	my $quelled  = $stats->{QUELLED};
	my $notices  = $stats->{NOTICES};
	my $untested = $stats->{UNTESTED};
	my $fips     = $stats->{FIPS};
	my $tests_expected   = $stats->{TESTS_EXPECTED};
	my $testsuite_errors = $stats->{TESTSUITE_ERRORS};
	
	my $res = "";
	
	$res .= '<h4>Journal statistics:</h4>'."\n";
	$res .= '<table class="info">'."\n";
	
	if (   defined $tests_expected && defined $tests_total 
		&& $tests_expected != $tests_total )
	{
		$res .= info_line('Tests Expected', $tests_expected, $fluorescent_yellow );
	}
	
	$res .= info_line('Tests Total', $tests_total ) if defined $tests_total;
	
	if ( defined $tests_passed ) { # zero value is printed too
		$res .= info_line('Tests Passed', $tests_passed, $fluorescent_green )
	}
	$res .= info_line('Tests Failed',  $failures, $fluorescent_red ) if $failures;
	$res .= info_line('Test suite errors',  $testsuite_errors, $fluorescent_red ) if $testsuite_errors;
	$res .= info_line('Warnings',  $warnings, $fluorescent_yellow ) if $warnings;
	$res .= info_line('Quelled problems',  $quelled ) if $quelled;
	$res .= info_line('Tests not relevant/not applicable',  $untested ) if $untested;
	$res .= info_line('Notices',  $notices ) if $notices;
	$res .= info_line('FIPs',  $fips ) if $fips;
	
	$res .= '</table>'."\n";
	
	return $res;
}
#----------------------------------------------------------------------

sub guess_journal_type {
	my ($filename) = @_;
	
	if ( $filename =~ /\.journal$/ ) {
		return "TET";
	}
	elsif ( $filename =~ /\.pythonjournal$/ ) {
		return "PYTHON";
	}
	elsif ( $filename =~ /\.tcljournal$/ ) {
		return "TCL";
	}
	elsif ( $filename =~ /\.simplejournal$/ ) {
		return "SIMPLE";
	}
	
	return undef;
}
#-----------------------------------------------------------------------

1; # Return value
